#
# Copyright (c) 2006-2009 National ICT Australia (NICTA), Australia
#
# Copyright (c) 2004-2009 WINLAB, Rutgers University, USA
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# = omfPubSubService.rb
#
# == Description
#
# This file implements a Publish/Subscribe Service Helper.
# This PubSub Service Helper is based on XMPP.
# This current implementation uses the library XMPP4R.
#
#

require "xmpp4r"
require "xmpp4r/pubsub"
require "xmpp4r/pubsub/helper/servicehelper"
require 'omf-common/mobject'
#Jabber::debug = true

#
# This class subclasses 'Jabber::PubSub::ServiceHelper' because its 
# 'unsubscribe_from' method is broken. Indeed, as stated in the XMPP4R v0.4 API, 
# it does NOT support the 'subid' field. However, the OpenFire v3.6 server (which
# we currently use as XMPP Server) requires the use of that field to process 
# unsubsribe requests, otherwise it replies with a 'Bad request' error.
#
class MyServiceHelper < Jabber::PubSub::ServiceHelper
  #
  # Perform a 'unsubscribe_from' from scratch
  #
  def unsubscribe_from_fixed (node,subid)
    iq = basic_pubsub_query(:set)
    sub = REXML::Element.new('unsubscribe')
    sub.attributes['node'] = node
    sub.attributes['jid'] = @stream.jid.strip.to_s
    sub.attributes['subid']=subid
    iq.pubsub.add(sub)
    ret = false
    @stream.send_with_id(iq) do |reply|
      ret = reply.kind_of?(Jabber::Iq) and reply.type == :result
    end # @stream.send_with_id(iq)
    ret  
  end
  
  #
  # Send a ping to the PubSub server
  # implemented according to
  # http://xmpp.org/extensions/xep-0199.html#c2s
  #
  def ping
    iq = Jabber::Iq.new(:get, @stream.jid.domain)
    iq.from = @stream.jid
    ping = iq.add(REXML::Element.new('ping'))
    ping.add_namespace 'urn:xmpp:ping'
    @stream.send_with_id(iq) do |reply|
      ret = reply.kind_of?(Jabber::Iq) and reply.type == :result
    end # @stream.send_with_id(iq)
  end

  #
  # Another bug in XMPP4r
  # owner field is missing when purging items
  # see http://www.ejabberd.im/node/3413
  #
  def purge_items_from(node)
    iq = basic_pubsub_query(:set,true)
    purge = REXML::Element.new('purge')
    purge.attributes['node'] = node
    iq.pubsub.add(purge)
    @stream.send_with_id(iq)
  end
  
end

# 
# This class defines a PubSub Service Helper for OMF entities.
# Following the PubSub communication paradigm, this class will be the
# 'XMPP Client' which will talk to an 'XMPP Server'. Currently, we are 
# using OpenFire v3.6 as the 'XMPP Server'.
# This Service Helper will handle all messages to and from the XMPP Server.
#
class OmfPubSubService < MObject

  #
  # Create a new instance of PubSub Service Helper
  # (JID = Jabber ID)
  # 
  # - user = [String] username to use to connect to the PubSub Server  
  # - password = [String], password to use for this PubSub client
  # - host = [String] hostname (or IP address) of the PubSub Server   
  #
  def initialize(user, password, host)
    # Set internal attributes
    @userJID = "#{user}@#{host}"
    @password = password
    @host = host
    @pubsubjid = "pubsub.#{host}"
    @connecting = false
    
    # First open a connection for this Helper to interact with the PubSub Server
    @clientHelper = Jabber::Client.new(@userJID)
    connect(@clientHelper)
    # when the other side closes the connection, try to reconnect after 10s
    @clientHelper.on_exception { sleep 10; connect(@clientHelper) }
    
    # remove subscriptions from previous experiments
    leave_all_pubsub_nodes()   
  end
  
  #
  # connect to the XMPP server
  # cl - global Jabber::Client instance
  #
  def connect(cl)
    # the exception handler may call this function multiple times
    # so don't do anything if we are already waiting for a connection
    return if @connecting
    @connecting = true
    # we are passing the hostname here to prevent xmpp4r from trying to resolve
    # the DNS SRV record
    debug "Connecting to XMPP server: '#{@host}' ..."
    begin
      cl.connect(@host)
    rescue
      debug "Could not connect to XMPP server: '#{@host}', retrying in 30s"
      sleep 30
      retry
    end

    # we are connected, now register a user
    begin
      cl.register(@password)
      # if the user already exists, we receive an error 409 ("conflict: ") and ignore it
      # otherwise we report it
    rescue Exception => ex
      if ("#{ex}" != "conflict: ")
        then raise "OmfPubSubService - Failed to register user #{@userJID} - Error: '#{ex}'"
      end
    end
    cl.auth(@password)
    cl.send(Jabber::Presence.new)
    @service = MyServiceHelper.new(cl, @pubsubjid)
    debug "Connection established to XMPP server: '#{@host}'"
    @connecting = false
  end

  #
  # Set callback for incoming message (i.e. "PubSub events")
  #
  # - &bock = the block of commands that will process the message
  #
  def add_event_callback (&block)
    begin
      @service.add_event_callback(&block)
    rescue Exception => ex
      raise "OmfPubSubService - add_event_callback - Error registering callback - '#{ex}'"
    end
  end

  #
  # Create a PubSub node (aka group, or discussion board)
  # Do nothing if the PubSub node already exists
  # Openfire automatically adds the creator of the node
  # to the subscriber list and does not allow it to
  # unsubscribe. Ejabberd does not do that. To stay
  # compatible with openfire, we replicate its behaviour
  # by subscribing to the node we create.
  #
  # - node = [String] name of the node to create
  #
  # [Return] True/False
  #
  def create_pubsub_node(node)
    begin
      @service.create_node(node,Jabber::PubSub::NodeConfig.new(nil,{
        "pubsub#title" => "#{node}",
        "pubsub#node_type" => "leaf",
        "pubsub#send_item_subscribe" => "1",
        "pubsub#max_items" => "1",
        "pubsub#persist_items" => "1",
        "pubsub#notify_retract" => "0",
        "pubsub#send_last_published_item" => "on_sub",
        "pubsub#publish_model" => "open"}))
      rescue Exception => ex
        # if the node exists ("conflict"), do not raise an exception
        if ("#{ex}"=="conflict: ")
          # delete any existing items first
          purge_pubsub_node(node)
        else
          raise "OmfPubSubService - create_pubsub_node - Unhandled Exception - '#{ex}'"
        end
      end
      # openfire subscribes us automatically, so this is just for ejabberd:
      join_pubsub_node(node)
      return true
    end

    #
    # Publish a new message (item) to a PubSub node
    #
    # - node = [String] name of the PubSub node 
    # - item = [Jabber::item] the PubSub item to publish
    #
    def publish_to_node(node,item)
      begin
        @service.publish_item_to(node,item)
      rescue Exception => ex
        if ("#{ex}"=="item-not-found: ")
          debug "publish_to_node - Node '#{node}' does not exist!"
          return false
        end
        if ("#{ex}"=="forbidden: ")
          debug "publish_to_node - Not allowed to publish to '#{node}'!"
          return false
        end	
        raise "OmfPubSubService - publish_to_node - Unhandled Exception - '#{ex}'"
      end
    end

    #
    # Subscribe to a PubSub node
    # Do nothing if the node does not exist or if we are already subscribed
    #
    # - node = [String] name of the PubSub node 
    #
    def join_pubsub_node(node)
      #debug "TDEBUG - join_pubsub_node - called with: '#{node}'"
      begin
        @service.subscribe_to(node)
      rescue Exception => ex
        if ("#{ex}"=="item-not-found: ")
          debug "join_pubsub_node - Node '#{node}' does not exist!"
          return false
        end
        raise "OmfPubSubService - join_pubsub_node - Unhandled Exception - '#{ex}'"
      end
      return true
    end

    #
    # Unsubscribe from a given PubSub node
    #
    # - node = [String] name of the PubSub node 
    # - subid = the subscription ID for this PubSub node
    #
    def leave_pubsub_node(node, subid)
      begin
        @service.unsubscribe_from_fixed(node, subid)
      rescue Exception => ex
        if ("#{ex}"=="item-not-found: ")
          debug "leave_pubsub_node - Node '#{node}' does not exist!"
          return true
        end
	if ("#{ex}"=="unexpected-request: ")
          return true
        end
        raise "OmfPubSubService - leave_pubsub_node - Unhandled Exception - '#{ex}'"
      end
      return true
    end
    
    #
    # Unsubscribe from all PubSub nodes
    #
    def leave_all_pubsub_nodes()
      listAllSubscription = get_all_pubsub_subscriptions
      #debug "TDEBUG - List BEFORE Leaving all: #{listAllSubscription}"
      listAllSubscription.each { |sub|
        leave_pubsub_node(sub.node, sub.subid)
      }
      #debug "TDEBUG - LIST AFTER Leaving - #{get_all_pubsub_subscriptions}"
    end

    #
    # Unsubscribe from all PubSub nodes except the ones containing the substring 'prefix'
    #
    def leave_all_pubsub_nodes_except(prefix)
      listAllSubscription = get_all_pubsub_subscriptions
      #debug "TDEBUG - List BEFORE Leaving all except the ones containing '#{prefix}: #{listAllSubscription}'"
      listAllSubscription.each { |sub|
        if (!sub.node.include? prefix) 
          then leave_pubsub_node(sub.node, sub.subid)
        end
      }
      #debug "TDEBUG - LIST AFTER Leaving - #{get_all_pubsub_subscriptions}"
    end

    #
    # Remove a PubSub node (aka group, or discussion board)
    # Do nothing if the PubSub node doesn't exist
    #
    # - node = [String] name of the node to remove
    #
    def remove_pubsub_node(node)
      begin
        # ejabberd keeps items even when the node has been deleted
        # so delete them here first
        purge_pubsub_node(node)
        @service.delete_node(node)
      rescue Exception => ex
        if ("#{ex}"!="item-not-found: ") 
          raise "OmfPubSubService - remove_pubsub_node - Unhandled Exception - '#{ex}'"
        end
      end
    end

    #
    # Unsubscribe and remove all PubSub nodes currently subscribed to
    #
    def remove_all_pubsub_nodes()
      listAllSubscriptions = get_all_pubsub_subscriptions
      listAllSubscriptions.each { |sub|
        leave_pubsub_node(sub.node, sub.subid)
        remove_pubsub_node(sub.node)
      }
    end

    #
    # Leave all PubSub nodes currently subscribed to
    # Delete the PubSub user
    # Stop the exception handler
    # Close the connection to the PubSub server
    #
    def quit()
      debug "Removing user from PubSub server and closing connection"
      remove_all_pubsub_nodes
      @clientHelper.remove_registration
      @clientHelper.on_exception {}
      @clientHelper.close
    end

    #
    # Purge all items from a given PubSub node
    #
    # - node = [String] name of the PubSub node 
    #
    def purge_pubsub_node(node)
      begin
        @service.purge_items_from(node)
      rescue Exception => ex
        if ("#{ex}"!="forbidden: " && "#{ex}"!="item-not-found: ")
          raise "OmfPubSubService - purge_pubsub_node - Unhandled Exception - '#{ex}'"
        end
      end
    end
    
    #
    # Returns all subscriptions for this PubSub Service Helper
    #
    # [Return] Hash of Strings
    #
    def get_all_pubsub_subscriptions
      list = []
      begin
        list = @service.get_subscriptions_from_all_nodes
      rescue Exception => ex
        raise "OmfPubSubService - get_all_pubsub_subscriptions - ERROR - '#{ex}'"
      end
      list
    end
    
    #
    # Send a ping to the PubSub server
    #
    def ping
      begin
        @service.ping
      rescue Exception => ex
        raise "OmfPubSubService - ping - ERROR - '#{ex}'"
      end
    end
    
  end #class
